""" Specify callable packages from pyscnomics """

from . import tools
from . import econ
from . import contracts
from . import dataset
from . import io
from . import optimize

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("pyscnomics")
except PackageNotFoundError:
    __version__ = "unknown"
